#include "stm32f0xx_hal.h"
#include "lcd1803.h"

/*
 *  Private functions
 */
#define LCD1803_Enable_IS(disp)   LCD1803_CMD(disp->spi, disp->Reg_FuncSet1|0b00000001);  // Enable IS
#define LCD1803_Disable_IS(disp)  LCD1803_CMD(disp->spi, disp->Reg_FuncSet1);  // Disable IS
#define LCD1803_Enable_RE(disp)   LCD1803_CMD(disp->spi, disp->Reg_FuncSet2); // Enable RE bit
#define LCD1803_Disable_RE(disp)  LCD1803_CMD(disp->spi, disp->Reg_FuncSet1); // Disable RE bit

static void LCD1803_CMD(SPI_HandleTypeDef *spi, uint8_t cmd);   // send command & wait for finish
static void LCD1803_WriteI(SPI_HandleTypeDef *spi, uint8_t b);  // write instruction
static uint8_t LCD1803_ReadS(SPI_HandleTypeDef *spi);           // read status
static void LCD1803_WriteB(SPI_HandleTypeDef *spi, uint8_t b);      // write byte
static uint8_t LCD1803_ReadB(SPI_HandleTypeDef *spi);               // read bvte

/*
 *  Functions
 */

/** \brief Initialize display connected to given spi port
 *
 * \param disp  Ptr to display struct
 * \param spi   Ptr to SPI handle
 *
 *  It is assumed that all SPI and GPIO settings already have been done by
 *  user. SPI port should be set to bidirectional mode, cpol=1, cpha=1.
 *  Display is cleared and initialized to normal font size.
 */
void LCD1803_Init(LCD1803_t *disp, SPI_HandleTypeDef *spi) {

    disp->spi = spi;

    /*  Initialize backup registers with function codes and default values*/

    disp->Reg_DH        = LCD1803_FC_DH|0x0c;       // POR: UD1+UD2 set
    disp->Reg_DispCur   = LCD1803_FC_DISPCONTROL;
    disp->Reg_EntryMode = LCD1803_FC_ENTRYMODE;
    disp->Reg_ExtFunc   = LCD1803_FC_EXTFUNC|0x01;  // NW bit set for 4 line display
    disp->Reg_FuncSet1  = LCD1803_FC_FUNCSET|0x18;  // 8-bit data mode, N bit set
    disp->Reg_FuncSet2  = LCD1803_FC_FUNCSET|0x1a;  // same as above, but RE enabled
    disp->Reg_Osc       = LCD1803_FC_OSC;
    disp->Reg_Power     = LCD1803_FC_POWER|0x0c;    // Ion bit set, Bon bit set

    /* Now init display */

    LCD1803_Set_OscFreq(disp, LCD1803_FREQ_540);
    LCD1803_Set_Contrast(disp, 0x0f);
    LCD1803_Enable_IS(disp);
    LCD1803_CMD(disp->spi, LCD1803_FC_FOLLOWER|0x0e); // Divider on
    LCD1803_Disable_IS(disp);
    LCD1803_Clear(disp);
    LCD1803_Set_DisplayOn(disp, 1);
    LCD1803_Set_EntryMode(disp, LCD1803_ENTRYMODE_INC);
}


/** \brief DeInitialize display
 *
 * \param disp  Ptr to display struct
 *
 */
void LCD1803_DeInit(LCD1803_t *disp) {
    LCD1803_Set_DisplayOn(disp, 0);
    LCD1803_Enable_IS(disp);
    LCD1803_CMD(disp->spi, LCD1803_FC_FOLLOWER);
    LCD1803_CMD(disp->spi, LCD1803_FC_POWER);
    LCD1803_Disable_IS(disp);
}


/** \brief Set bidirect mode
 *
 * \param disp  Ptr to display struct
 * \param flags Specifies horizontal/vertical bidirect mode
 *
 */
void LCD1803_Set_BidirectMode(LCD1803_t *disp, uint8_t flags) {
    LCD1803_Enable_RE(disp);
    LCD1803_CMD(disp->spi, LCD1803_FC_BIDIRECT|(flags&0x03));
    LCD1803_Disable_RE(disp);
}


/** \brief Set entry mode
 *
 * \param disp  Ptr to display struct
 * \param mode  Entry mode
 *
 */
void LCD1803_Set_EntryMode(LCD1803_t *disp, LCD1803_EntryMode_t mode) {
    switch(mode) {
    case LCD1803_ENTRYMODE_DEC:
        disp->Reg_EntryMode &= ~0x02;
        break;

    case LCD1803_ENTRYMODE_INC:
        disp->Reg_EntryMode |= 0x02;
        break;
    }

    LCD1803_CMD(disp->spi, disp->Reg_EntryMode);
}


/** \brief Set display on
 *
 * \param disp  Ptr to display struct
 * \param flag  If true enable display, otherwise disable
 *
 */
void LCD1803_Set_DisplayOn(LCD1803_t *disp, uint8_t flag) {
    if(flag) {
        disp->Reg_DispCur |= 0x04;
    }
    else {
        disp->Reg_DispCur &= ~0x04;
    }
    LCD1803_CMD(disp->spi, disp->Reg_DispCur);
}


/** \brief Set cursor mode
 *
 * \param disp  Ptr to display struct
 * \param mode  Cursor mode
 *
 */
void LCD1803_Set_CursorMode(LCD1803_t *disp, LCD1803_CursorMode_t mode) {
    switch(mode) {
    case LCD1803_CURSORMODE_OFF:
        disp->Reg_DispCur &= ~0x03;
        disp->Reg_ExtFunc &= ~0x02;
        break;

    case LCD1803_CURSORMODE_ON:
        disp->Reg_DispCur = (disp->Reg_DispCur&~0x01)|0x02;
        disp->Reg_ExtFunc &= 0x02;
        break;

    case LCD1803_CURSORMODE_BLINK:
//        disp->Reg_DispCur |= 0x03;
        disp->Reg_DispCur = (disp->Reg_DispCur&~0x02)|0x01;
        disp->Reg_ExtFunc &= ~0x02;
        break;

    case LCD1803_CURSORMODE_INV:
        disp->Reg_DispCur &= ~0x03;
        disp->Reg_ExtFunc |= 0x02;
        break;

    }

    LCD1803_CMD(disp->spi, disp->Reg_DispCur);
    LCD1803_Enable_RE(disp);
    LCD1803_CMD(disp->spi, disp->Reg_ExtFunc);
    LCD1803_Disable_RE(disp);
}


/** \brief Set shift on
 *
 * \param disp  Ptr to display struct
 * \param flag  If true enable shift
 *
 */
void LCD1803_Set_ShiftOn(LCD1803_t *disp, uint8_t flag) {
    if(flag) {
        disp->Reg_EntryMode |= 0x01;
    }
    else {
        disp->Reg_EntryMode &= ~0x01;
    }

    LCD1803_CMD(disp->spi, disp->Reg_EntryMode);
}


/** \brief Set reverse on
 *
 * \param disp  Ptr to display struct
 * \param flag  If true enable reverse display
 *
 */
void LCD1803_Set_ReverseOn(LCD1803_t *disp, uint8_t flag) {
    if(flag) {
        disp->Reg_FuncSet2 |= 0x01;
    }
    else {
        disp->Reg_FuncSet2 &= ~0x01;
    }

    LCD1803_Enable_RE(disp);
    LCD1803_Disable_RE(disp);
}


/** \brief Set blinking of CGRAM on
 *
 * \param disp  Ptr to display struct
 * \param flag  If true enable blinking
 *
 */
void LCD1803_Set_BlinkOn(LCD1803_t *disp, uint8_t flag) {
    if(flag) {
        disp->Reg_FuncSet2 |= 0x04;
    }
    else {
        disp->Reg_FuncSet2 &= ~0x04;
    }

    LCD1803_Enable_RE(disp);
    LCD1803_Disable_RE(disp);
}


/** \brief Set double height mode
 *
 * \param disp  Ptr to display struct
 * \param mode  DH mode
 *
 */
void LCD1803_Set_DHMode(LCD1803_t *disp, LCD1803_DoubleHeight_t mode) {
    if(mode == LCD1803_DH_NONE) {
        disp->Reg_FuncSet1 &= ~0x04;
    }
    else {
        disp->Reg_FuncSet1 |= 0x04;
        disp->Reg_DH = (disp->Reg_DH&~0x0c)|((mode-1)<<2);
    }

    LCD1803_Enable_RE(disp);
    LCD1803_CMD(disp->spi, disp->Reg_DH);
    LCD1803_Disable_RE(disp);
}


/** \brief Set shift/scroll enable
 *
 * \param disp  Ptr to display struct
 * \param mode  Shift mode (shift, dot)
 * \param mask  Each line has a bit (0...)
 *
 */
void LCD1803_Set_ShiftMode(LCD1803_t *disp, LCD1803_ShiftMode_t mode, uint8_t mask) {

    LCD1803_Enable_RE(disp);
    switch(mode) {
    case LCD1803_SHIFTMODE_DOT:
        disp->Reg_DH = (disp->Reg_DH&~0x01);
        break;

    case LCD1803_SHIFTMODE_NORM:
        disp->Reg_DH = disp->Reg_DH|0x01;
        break;
    }

    LCD1803_CMD(disp->spi, disp->Reg_DH);
    LCD1803_Disable_RE(disp);

    LCD1803_Enable_IS(disp);
    LCD1803_Enable_RE(disp);
    LCD1803_CMD(disp->spi, LCD1803_FC_SHIFTENABLE|(mask&0x0f));
    LCD1803_Disable_RE(disp);   // also disables IS
}


/** \brief Set contrast
 *
 * \param disp      Ptr to display struct
 * \param contrast  0..31
 *
 */
void LCD1803_Set_Contrast(LCD1803_t *disp, uint8_t contrast) {
    LCD1803_Enable_IS(disp);
    LCD1803_CMD(disp->spi, LCD1803_FC_CONTRAST|(contrast&0x0f));
    disp->Reg_Power = (disp->Reg_Power&~0x03)|((contrast>>4)&0x03);
    LCD1803_CMD(disp->spi, disp->Reg_Power);
    LCD1803_Disable_IS(disp);
}


/** \brief Set character ROM
 *
 * \param disp  Ptr to display struct
 * \param rom   Rom to use
 *
 */
void LCD1803_Set_Rom(LCD1803_t *disp, LCD1803_CharRom_t rom) {
    LCD1803_Enable_RE(disp);
    LCD1803_CMD(disp->spi, LCD1803_FC_ROMSEL);
    LCD1803_WriteB(disp->spi, rom<<2);
    LCD1803_Disable_RE(disp);
}


/** \brief Set bias
 *
 * \param disp  Ptr to display struct
 * \param bias  0..3
 *
 */
void LCD1803_Set_Bias(LCD1803_t *disp, LCD1803_Bias_t bias) {
    LCD1803_Enable_IS(disp);
    disp->Reg_Osc = (disp->Reg_Osc&~0x08)|((bias&0x01)<<3);
    LCD1803_CMD(disp->spi, disp->Reg_Osc);
    LCD1803_Disable_IS(disp);
    LCD1803_Enable_RE(disp);
    disp->Reg_DH = (disp->Reg_DH&~0x02)|(bias&0x02);
    LCD1803_CMD(disp->spi, disp->Reg_DH);
    LCD1803_Disable_RE(disp);
}


/** \brief Set osc. freq
 *
 * \param disp  Ptr to display struct
 * \param freq  Frequency
 *
 */
void LCD1803_Set_OscFreq(LCD1803_t *disp, LCD1803_OscFreq_t freq) {
    LCD1803_Enable_IS(disp);
    disp->Reg_Osc = (disp->Reg_Osc&~0x07)|freq;
    LCD1803_CMD(disp->spi, disp->Reg_Osc);
    LCD1803_Disable_IS(disp);
}


/** \brief Set temp coeff.
 *
 * \param disp  Ptr to display struct
 * \param tc    Temp. coefficient
 *
 */
void LCD1803_Set_TC(LCD1803_t *disp, uint8_t tc) {
    LCD1803_Enable_RE(disp);
    LCD1803_CMD(disp->spi, LCD1803_FC_TC);
    LCD1803_WriteB(disp->spi, tc);
    while(LCD1803_ReadS(disp->spi)&LCD1803_FLAG_BUSY);     // wait until BF cleared
    LCD1803_Disable_RE(disp);
}


/** \brief Set CGRAM address
 *
 * \param disp  Ptr to display struct
 * \param addr  Address
 */
void LCD1803_CGRAMADDR(LCD1803_t *disp, uint8_t addr) {
    LCD1803_CMD(disp->spi, LCD1803_FC_CGRAMDDR|(addr&0x3f));
}


/** \brief Set SEGRAM address
 *
 * \param disp  Ptr to display struct
 * \param addr  Address
 */
void LCD1803_SEGRAMADDR(LCD1803_t *disp, uint8_t addr) {
    LCD1803_Enable_IS(disp);
    LCD1803_CMD(disp->spi, LCD1803_FC_SEGRAMADDR|(addr&0x1f));
    LCD1803_Disable_IS(disp);
}


/** \brief Clears display and set cursor to home
 *
 * \param disp  Ptr to display struct
 *
 */
void LCD1803_Clear(LCD1803_t *disp) {
    LCD1803_CMD(disp->spi, LCD1803_FC_CLEAR);
}


/** \brief Set cursor to home position
 *
 * \param disp  Ptr to display struct
 *
 */
void LCD1803_Home(LCD1803_t *disp) {
    LCD1803_CMD(disp->spi, LCD1803_FC_HOME);
}


/** \brief Shift cursor
 *
 * \param disp  Ptr to display struct
 * \param dir   Shift direction
 *
 */
void LCD1803_ShiftCursor(LCD1803_t *disp, LCD1803_ShiftDirection_t dir) {
    switch(dir) {
    case LCD1803_SHIFT_LEFT:
        LCD1803_CMD(disp->spi, LCD1803_FC_SHIFT);
        break;

    case LCD1803_SHIFT_RIGHT:
        LCD1803_CMD(disp->spi, LCD1803_FC_SHIFT|0x04);
        break;
    }
}


/** \brief Shift display
 *
 * \param disp  Ptr to display struct
 * \param dir   Shift direction
 *
 *  Lines to shift need to be enabled first
 */
void LCD1803_ShiftDisplay(LCD1803_t *disp, LCD1803_ShiftDirection_t dir) {
    switch(dir) {
    case LCD1803_SHIFT_LEFT:
        LCD1803_CMD(disp->spi, LCD1803_FC_SHIFT|0x08);
        break;

    case LCD1803_SHIFT_RIGHT:
        LCD1803_CMD(disp->spi, LCD1803_FC_SHIFT|0x0c);
        break;
    }
}


/** \brief Scroll display vertically by dots
 *
 * \param disp  Ptr to display struct
 * \param dots  Number of dots
 */
void LCD1803_ScrollDisplay(LCD1803_t *disp, uint8_t dots) {
    LCD1803_Enable_RE(disp);
    LCD1803_CMD(disp->spi, LCD1803_FC_SCROLL|(dots&0x3f));
    LCD1803_Disable_RE(disp);
}


/** \brief Set DDRAM address
 *
 * \param disp  Ptr to display struct
 * \param addr  Address
 */
void LCD1803_DDRAMADDR(LCD1803_t *disp, uint8_t addr) {
    LCD1803_CMD(disp->spi, LCD1803_FC_DDRAMADDR|addr);
}


/** \brief Write data
 *
 * \param disp  Ptr to display struct
 * \param data  Ptr to data
 * \param size  Size of data
 */
void LCD1803_Write(LCD1803_t *disp, const uint8_t *data, int size) {
    uint8_t b = LCD1803_STARTBYTE|LCD1803_FLAG_RS;

    HAL_SPI_Transmit(disp->spi, &b, 1, HAL_MAX_DELAY);

    while(size-- > 0) {
        b = *data&0x0f;
        HAL_SPI_Transmit(disp->spi, &b, 1, HAL_MAX_DELAY);
        b = *data>>4;
        HAL_SPI_Transmit(disp->spi, &b, 1, HAL_MAX_DELAY);
        data++;
    }
}


/** \brief Move cursor to X/Y position
 *
 * \param disp  Ptr to display struct
 * \param x     X position starting from 0
 * \param y     y position starting from 0
 */
void LCD1803_MoveXY(LCD1803_t *disp, uint8_t x, uint8_t y) {
    LCD1803_DDRAMADDR(disp, x+y*LCD1803_ROWOFFSET);
}



/*
 *  Private functions used for SPI I/O
 */

static void LCD1803_CMD(SPI_HandleTypeDef *spi, uint8_t inst) {
    LCD1803_WriteI(spi, inst);          // send command
    while(LCD1803_ReadS(spi)&LCD1803_FLAG_BUSY);     // wait until BF cleared
}

static void LCD1803_WriteI(SPI_HandleTypeDef *spi, uint8_t inst) {
    uint8_t tx[3];

    tx[0] = LCD1803_STARTBYTE;
    tx[1] = inst&0x0f;
    tx[2] = inst>>4;

    HAL_SPI_Transmit(spi, tx, 3, HAL_MAX_DELAY);
}

static uint8_t LCD1803_ReadS(SPI_HandleTypeDef *spi) {
    uint8_t tx = LCD1803_STARTBYTE|LCD1803_FLAG_RW, rx;

    HAL_SPI_Transmit(spi, &tx, 1, HAL_MAX_DELAY);

    tx = 0;
    HAL_SPIEx_FlushRxFifo(spi);
    HAL_SPI_TransmitReceive(spi, &tx, &rx, 1, HAL_MAX_DELAY);

    return rx;
}

static void LCD1803_WriteB(SPI_HandleTypeDef *spi, uint8_t b) {
    uint8_t tx[3];

    tx[0] = LCD1803_STARTBYTE|LCD1803_FLAG_RS;
    tx[1] = b&0x0f;
    tx[2] = b>>4;

    HAL_SPI_Transmit(spi, tx, 3, HAL_MAX_DELAY);
}

static uint8_t LCD1803_ReadB(SPI_HandleTypeDef *spi) {
    uint8_t b = LCD1803_STARTBYTE|LCD1803_FLAG_RS|LCD1803_FLAG_RW;
    uint8_t zero = 0;

    HAL_SPI_Transmit(spi, &b, 1, HAL_MAX_DELAY);
    HAL_SPIEx_FlushRxFifo(spi);
    HAL_SPI_TransmitReceive(spi, &zero, &b, 1, HAL_MAX_DELAY);

    return b;
}
