#ifndef LCD1803_H_INCLUDED
#define LCD1803_H_INCLUDED

#include "stm32f0xx_hal.h"

/*
 *  Some defs
 */

#define LCD1803_STARTBYTE   0b00011111  // Startbyte of serial transfer
#define LCD1803_FLAG_RW     0b00100000  // RW flag
#define LCD1803_FLAG_RS     0b01000000  // RS flag
#define LCD1803_FLAG_BUSY   0b10000000  // Busy flag
#define LCD1803_ROWOFFSET   0x20        // Addr offset per line

/*  Display function codes */

#define LCD1803_FC_CLEAR        0b00000001  // clear display
#define LCD1803_FC_HOME         0b00000010  // home
#define LCD1803_FC_POWERDOWN    0b00000010  // IS=1: power down
#define LCD1803_FC_ENTRYMODE    0b00000100  // set entry mode
#define LCD1803_FC_BIDIRECT     0b00000100  // RE=1: bidirection mode
#define LCD1803_FC_DISPCONTROL  0b00001000  // Display mode
#define LCD1803_FC_EXTFUNC      0b00001000  // RE=1: extended functions
#define LCD1803_FC_SHIFT        0b00010000  // cursor/display shift
#define LCD1803_FC_DH           0b00010000  // RE=1: double height control
#define LCD1803_FC_OSC          0b00010000  // IS=1: Osc freq
#define LCD1803_FC_SHIFTENABLE  0b00010000  // RE=1, IS=1: shift/scroll enable
#define LCD1803_FC_FUNCSET      0b00100000  // function set
#define LCD1803_FC_CGRAMDDR     0b01000000  // set CGRAM address
#define LCD1803_FC_SEGRAMADDR   0b01000000  // IS=1: set SEGRAM address
#define LCD1803_FC_POWER        0b01010000  // IS=1: power control
#define LCD1803_FC_FOLLOWER     0b01100000  // IS=1: follower control
#define LCD1803_FC_CONTRAST     0b01110000  // IS=1: contrast set
#define LCD1803_FC_DDRAMADDR    0b10000000  // set DDRAM addr
#define LCD1803_FC_SCROLL       0b10000000  // RE=1: set scroll quantity
#define LCD1803_FC_TC           0b01110110  // RE=1: set temp coeff
#define LCD1803_FC_ROMSEL       0b01110010  // RE=1: select char rom


/*
 *  Struct used to handle display instance
 */
typedef struct {
    SPI_HandleTypeDef   *spi;           // Ptr to SPI
    uint8_t             Reg_EntryMode;  // Backup of EntryMode reg
    uint8_t             Reg_DispCur;    // Display / Cursor mode reg backup
    uint8_t             Reg_ExtFunc;    // Extended func reg backup
    uint8_t             Reg_DH;         // DH reg backup
    uint8_t             Reg_Osc;        // Osc reg backup
    uint8_t             Reg_FuncSet1;   // FuncSet1 reg backup
    uint8_t             Reg_FuncSet2;   // Funcset2 reg backup
    uint8_t             Reg_Power;      // Power reg backup
} LCD1803_t;



/*
 *  Bits in Bidirect mode mask
 */
 #define    LCD1803_BIDIRECTMODE_HORIZ  0x01    // horizontal
 #define    LCD1803_BIDIRECTMODE_VERT   0x02    // vertical

/*
 *  Entry modes
 */
typedef enum {
    LCD1803_ENTRYMODE_INC,              // Increase cursor pos after write
    LCD1803_ENTRYMODE_DEC               // Decrease cursor pos after write
} LCD1803_EntryMode_t;

/*
 *  Cursor modes
 */
typedef enum  {
    LCD1803_CURSORMODE_OFF,             // Cursor off
    LCD1803_CURSORMODE_ON,              // Cursor on (line)
    LCD1803_CURSORMODE_BLINK,           // Cursor blinking
    LCD1803_CURSORMODE_INV              // Cursor inverting
} LCD1803_CursorMode_t;

/*
 *  Shift direction
 */
typedef enum {
    LCD1803_SHIFT_LEFT,                 // Shift left
    LCD1803_SHIFT_RIGHT                 // Shift right
} LCD1803_ShiftDirection_t;

/*
 *  Shift mode
 */
typedef enum {
    LCD1803_SHIFTMODE_NORM,             // Normal shift
    LCD1803_SHIFTMODE_DOT               // Dot scroll
} LCD1803_ShiftMode_t;

/*
 *  Double high modes
 */
typedef enum {
    LCD1803_DH_NONE,                    // No double height
    LCD1803_DH_NND,                     // Normal, Normal, Double
    LCD1803_DH_NDN,                     // Normal, Double, Normal
    LCD1803_DH_DD,                      // Double, Double
    LCD1803_DH_DNN                      // Double, Normal, Normal
} LCD1803_DoubleHeight_t;

/*
 *  Character roms
 */
typedef enum {
    LCD1803_ROM_A, LCD1803_ROM_B, LCD1803_ROM_C
} LCD1803_CharRom_t;

/*
 *  Bias modes
 */
typedef enum {
    LCD1803_BIAS_15,                    // 1/5
    LCD1803_BIAS_14,                    // 1/4
    LCD1803_BIAS_17,                    // 1/7
    LCD1803_BIAS_16                     // 1/6
} LCD1803_Bias_t;

/*
 *  Oscillator freq
 */
typedef enum {
    LCD1803_FREQ_420, LCD1803_FREQ_460, LCD1803_FREQ_500, LCD1803_FREQ_540,
    LCD1803_FREQ_580, LCD1803_FREQ_620, LCD1803_FREQ_640, LCD1803_FREQ_680
} LCD1803_OscFreq_t;


/*
 *  Functions
 */

/** \brief Initialize display connected to given spi port
 *
 * \param disp  Ptr to display struct
 * \param spi   Ptr to SPI handle
 *
 *  It is assumed that all SPI and GPIO settings already have been done by
 *  user. SPI port should be set to bidirectional mode, cpol=1, cpha=1.
 *  Display is cleared and initialized to normal font size.
 */
void LCD1803_Init(LCD1803_t *disp, SPI_HandleTypeDef *spi);


/** \brief DeInitialize display
 *
 * \param disp  Ptr to display struct
 *
 */
void LCD1803_DeInit(LCD1803_t *disp);


/** \brief Set bidirect mode
 *
 * \param disp  Ptr to display struct
 * \param flags Specifies horizontal/vertical bidirect mode
 *
 */
void LCD1803_Set_BidirectMode(LCD1803_t *disp, uint8_t flags);


/** \brief Set entry mode
 *
 * \param disp  Ptr to display struct
 * \param mode  Entry mode
 *
 */
void LCD1803_Set_EntryMode(LCD1803_t *disp, LCD1803_EntryMode_t mode);

/** \brief Set display on
 *
 * \param disp  Ptr to display struct
 * \param flag  If true enable display, otherwise disable
 *
 */
void LCD1803_Set_DisplayOn(LCD1803_t *disp, uint8_t flag);

/** \brief Set cursor mode
 *
 * \param disp  Ptr to display struct
 * \param mode  Cursor mode
 *
 */
void LCD1803_Set_CursorMode(LCD1803_t *disp, LCD1803_CursorMode_t mode);

/** \brief Set shift on
 *
 * \param disp  Ptr to display struct
 * \param flag  If true enable shift
 *
 */
void LCD1803_Set_ShiftOn(LCD1803_t *disp, uint8_t flag);

/** \brief Set reverse on
 *
 * \param disp  Ptr to display struct
 * \param flag  If true enable reverse disblay
 *
 */
void LCD1803_Set_ReverseOn(LCD1803_t *disp, uint8_t flag);

/** \brief Set blinking of CGRAM
 *
 * \param disp  Ptr to display struct
 * \param flag  If true enable blinking
 *
 */
void LCD1803_Set_BlinkOn(LCD1803_t *disp, uint8_t flag);

/** \brief Set double height mode
 *
 * \param disp  Ptr to display struct
 * \param mode  DH mode
 *
 */
void LCD1803_Set_DHMode(LCD1803_t *disp, LCD1803_DoubleHeight_t mode);

/** \brief Set shift/scroll enable
 *
 * \param disp  Ptr to display struct
 * \param mode  Shift mode (shift, dot)
 * \param mask  Each line has a bit (0...)
 *
 */
void LCD1803_Set_ShiftMode(LCD1803_t *disp, LCD1803_ShiftMode_t mode, uint8_t mask);

/** \brief Set contrast
 *
 * \param disp      Ptr to display struct
 * \param contrast  0..31
 *
 */
void LCD1803_Set_Contrast(LCD1803_t *disp, uint8_t contrast);

/** \brief Set character ROM
 *
 * \param disp  Ptr to display struct
 * \param rom   Rom to use
 *
 */
void LCD1803_Set_Rom(LCD1803_t *disp, LCD1803_CharRom_t rom);

/** \brief Set bias
 *
 * \param disp  Ptr to display struct
 * \param bias  0..3
 *
 */
void LCD1803_Set_Bias(LCD1803_t *disp, LCD1803_Bias_t bias);

/** \brief Set osc. freq
 *
 * \param disp  Ptr to display struct
 * \param freq  Frequency
 *
 */
void LCD1803_Set_OscFreq(LCD1803_t *disp, LCD1803_OscFreq_t freq);

/** \brief Set temp coeff.
 *
 * \param disp  Ptr to display struct
 * \param tc    Temp. coefficient
 *
 */
void LCD1803_Set_TC(LCD1803_t *disp, uint8_t tc);

/** \brief Set CGRAM address
 *
 * \param disp  Ptr to display struct
 * \param addr  Address
 */
void LCD1803_CGRAMADDR(LCD1803_t *disp, uint8_t addr);

/** \brief Set SEGRAM address
 *
 * \param disp  Ptr to display struct
 * \param addr  Address
 */
void LCD1803_SEGRAMADDR(LCD1803_t *disp, uint8_t addr);

/** \brief Clears display and set cursor to home
 *
 * \param disp  Ptr to display struct
 *
 */
void LCD1803_Clear(LCD1803_t *disp);

/** \brief Set cursor to home position
 *
 * \param disp  Ptr to display struct
 *
 */
void LCD1803_Home(LCD1803_t *disp);

/** \brief Shift cursor
 *
 * \param disp  Ptr to display struct
 * \param dir   Shift direction
 *
 */
void LCD1803_ShiftCursor(LCD1803_t *disp, LCD1803_ShiftDirection_t dir);

/** \brief Shift display
 *
 * \param disp  Ptr to display struct
 * \param dir   Shift direction
 *
 *  Lines to shift need to be enabled first
 */
void LCD1803_ShiftDisplay(LCD1803_t *disp, LCD1803_ShiftDirection_t dir);

/** \brief Scroll display vertically by dots
 *
 * \param disp  Ptr to display struct
 * \param dots  Number of dots
 */
void LCD1803_ScrollDisplay(LCD1803_t *disp, uint8_t dots);

/** \brief Set DDRAM address
 *
 * \param disp  Ptr to display struct
 * \param addr  Address
 */
void LCD1803_DDRAMADDR(LCD1803_t *disp, uint8_t addr);

/** \brief Move cursor to X/Y position
 *
 * \param disp  Ptr to display struct
 * \param x     X position starting from 0
 * \param y     y position starting from 0
 */
void LCD1803_MoveXY(LCD1803_t *disp, uint8_t x, uint8_t y);

/** \brief Write data
 *
 * \param disp  Ptr to display struct
 * \param data  Ptr to data
 * \param size  Size of data
 */
void LCD1803_Write(LCD1803_t *disp, const uint8_t *data, int size);

#endif /* LCD1803_H_INCLUDED */
